/* 80x86 disassembler for real mode and 16 bit segments.
 * 32 bit address and data prefixes are partially supported; there
 * may be some missing cases.
 *
 * Copyright 1993 by John F. Carr (jfc@athena.mit.edu)
 *
 */

#include <stdio.h>

#ifdef __GNUC__
#define INLINE inline
#else
#define INLINE
#endif

static const char *conditions[16] =
{
  "o", "no", "b", "ae", "e", "ne", "be", "a",
  "s", "ns", "p", "np", "le", "g", "le", "g"
};

static const char *cr[] =
{
  "cr0", "cr1", "cr2", "cr3", "cr4", "cr5", "cr6", "cr7"
};

static const char *dr[] =
{
  "dr0", "dr1", "dr2", "dr3", "dr4", "dr5", "dr6", "dr7"
};

static const char *tr[] =
{
  "tr0", "tr1", "tr2", "tr3", "tr4", "tr5", "tr6", "tr7"
};

static const char *freg[8] =
{
  "st(0)", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
};

static const char *farith[8] =
{
  "fadd", "fmul", "fcom", "fcomp", "fsub", "fsubr", "fdiv", "fdivr"
};

static const char *d9[8] =
{
  "flds", "fxch", "fsts", "fstps", "ftst/fxam", "fldc", "fmisc1", "fmisc2"
};

static const char *f00[8] =
{
  "sldt", "str", "lldt", "ltr", "verr", "verw", "???", "???"
};

static const char *f01[8] =
{
  "sgdt", "sidt", "lgdt", "lidt", "smsw", "???", "lmsw", "???"
};

static const char *shift[8] =
{
  "rol", "ror", "rcl", "rcr", "shl", "shr", "sal", "sar"
};

static const char *f7_code[8] =
{
  "test", "???", "not", "neg", "mul", "imul", "div", "idiv"
};

static const char *ff_code[8] =
{
  "inc", "dec", "call", "call far", "jmp", "jmp far", "push", "???"
};

static const char *arith[8] =
{
  "add", "or", "adc", "sbb", "and", "sub", "xor", "cmp"
};

static const char *reg8[8] =
{
  "al", "cl", "dl", "bl", "ah", "ch", "dh", "bh"
};

static const char *reg16[8] =
{
  "ax", "cx", "dx", "bx", "sp", "bp", "si", "di"
};

static const char *reg32[8] =
{
  "eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi"
};

static const char *sreg[8] =
{
  "es", "cs", "ss", "ds", "fs", "gs", "??", "??"
};

static const char *memref[8] =
{
  "bx+si", "bx+di", "bp+si", "bp+di", "si", "di", "bp", "bx"
};

static const char *fop[8] =
{
  "0", "1", "2", "3", "4", "5", "6", "7"
};

/* bit set: instruction takes explicit float register operand */
static unsigned char freg_op[8] =
{
  0xff, 0x03, 0xd3, 0x00, 0xf0, 0x35, 0xf0, 0x00
};

#if defined(i386) || defined(vax)
/* Little-endian machines which allow unaligned addresses.  */
#define immed16(code) *(unsigned short *)(code)
#define immed32(code) *(unsigned int *)(code)
#else
/* Big-endian or strict alignment.  */
#define immed16(code) code[0] | (code[1] << 8)
#define immed32(code) code[0] | (code[1] << 8) | (code[2] << 16) | (code[3] << 24)
#endif

#undef index

static INLINE const unsigned char *mod_rm(FILE *out, const unsigned char *code,
					  const char *seg, const char **regs,
					  int addr32)
{
  unsigned char rm = *code++;

  if (addr32)
    {
      char index[5];
      const char *scale;
      const char *base;

      if ((rm >> 6) == 3)
	{
	  fputs(regs[rm & 7], out);
	  return code;
	}
      if ((rm & 7) == 4)
	{
	  unsigned char sib = *code++;

	  base = reg32[sib & 7];

	  if (((sib >> 3) & 7) == 4)
	    {
	      index[0] = 0;
	      scale = "";
	    }
	  else
	    {
	      index[0] = '+';
	      index[1] = reg32[(sib >> 3) & 7][0];
	      index[2] = reg32[(sib >> 3) & 7][1];
	      index[3] = reg32[(sib >> 3) & 7][2];
	      index[4] = 0;

	      switch(sib >> 6)
		{
		case 0:
		  scale = "";
		  break;
		case 1:
		  scale = "*2";
		  break;
		case 2:
		  scale = "*4";
		  break;
		case 3:
		  scale = "*8";
		  break;
		}
	    }
	}
      else
	{
	  base = reg32[rm & 7];
	  index[0] = 0;
	  scale = "";
	}

      switch(rm >> 6)
	{
	case 2:
	  fprintf(out, "%s[%s%s%s%+d]", seg, reg32[rm & 7],
		  index, scale, immed32(code));
	  return code + 4;

	case 1:
	  fprintf(out, "%s[%s%s%s%+d]", seg, reg32[rm & 7], index, scale,
		  (signed char)code[0]);
	  return code + 1;

	case 0:
	  if ((rm & 7) == 5)
	    {
	      fprintf(out, "%s[%08x%s%s]", seg, immed32(code),
		      index, scale);
	      return code + 4;
	    }
	  fprintf(out, "%s[%s%s%s]", seg, reg32[rm & 7], index, scale);
	  return code;
	}
    }
  else
    {
      switch(rm >> 6)
	{
	case 3:
	  fputs(regs[rm & 7], out);
	  return code;

	case 2:
	  fprintf(out, "%s[%s%+d]", seg, memref[rm & 7], (short)immed16(code));
	  return code + 2;

	case 1:
	  fprintf(out, "%s[%s%+d]", seg, memref[rm & 7], (signed char)code[0]);
	  return code + 1;

	case 0:
	  if ((rm & 7) == 6)
	    {
	      fprintf(out, "%s[%04x]", seg, code[0] + (code[1] << 8));
	      return code + 2;
	    }
	  fprintf(out, "%s[%s]", seg, memref[rm & 7]);
	  return code;
	}
    }
}

static const unsigned char *mod_reg_rm(FILE *out, const unsigned char *code,
				   const char *seg, const char **sregs,
				   const char **regs, int addr32)
{
  fputs(sregs[(*code >> 3) & 7], out);
  putc(',', out);
  return mod_rm(out, code, seg, regs, addr32);
}

static const unsigned char *mod_reg_rm_r(FILE *out, const unsigned char *code,
				     const char *seg, const char **sregs,
				     const char **regs, int addr32)
{
  unsigned char rm = *code;

  code = mod_rm(out, code, seg, regs, addr32);
  putc(',', out);
  fputs(sregs[(rm >> 3) & 7], out);
  return code;
}

void dis_8086(FILE *out, unsigned int org,
	      register const unsigned char *code,
	      const unsigned char *ecode,
	      int def_size)
{
  const unsigned char *code0 = code - org;
  const char *seg = "";
  int data32 = def_size;
  int addr32 = def_size;
  int prefix = 0;
  while (code < ecode)
    {
      unsigned char opcode;
      register const char **wreg = data32 ? reg32 : reg16;
      if (! prefix)
	fprintf(out, "%04x:\t", code - code0);
      opcode = *code++;
      switch(opcode)
	{
	case 0x00:
	case 0x08:
	case 0x10:
	case 0x18:
	case 0x20:
	case 0x28:
	case 0x30:
	case 0x38:
	  fputs(arith[(opcode >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_reg_rm_r(out, code, seg, reg8, reg8, addr32);
	  break;

	case 0x01:
	case 0x09:
	case 0x11:
	case 0x19:
	case 0x21:
	case 0x29:
	case 0x31:
	case 0x39:
	  fputs(arith[(opcode >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_reg_rm_r(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0x02:
	case 0x0a:
	case 0x12:
	case 0x1a:
	case 0x22:
	case 0x2a:
	case 0x32:
	case 0x3a:
	  fputs(arith[(opcode >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_reg_rm(out, code, seg, reg8, reg8, addr32);
	  break;

	case 0x03:
	case 0x0b:
	case 0x13:
	case 0x1b:
	case 0x23:
	case 0x2b:
	case 0x33:
	case 0x3b:
	  fputs(arith[(opcode >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0x04:
	case 0x0c:
	case 0x14:
	case 0x1c:
	case 0x24:
	case 0x2c:
	case 0x34:
	case 0x3c:
	  fprintf(out, "%s al,%d", arith[opcode >> 3],
		  (signed char)*code++);
	  break;

	case 0x05:
	case 0x0d:
	case 0x15:
	case 0x1d:
	case 0x25:
	case 0x2d:
	case 0x35:
	case 0x3d:
	  if (data32)
	    {
	      fprintf(out, "%s eax,0x%x", arith[opcode >> 3], immed32(code));
	      code += 4;
	    }
	  else
	    {
	      fprintf(out, "%s ax,0x%hx", arith[opcode >> 3], immed16(code));
	      code += 2;
	    }
	  break;

	case 0x06:
	  fputs("push es", out);
	  break;

	case 0x07:
	  fputs("pop es", out);
	  break;

	case 0x0e:
	  fputs("push cs", out);
	  break;

	case 0x0f:
	  opcode = *code++;
	  switch (opcode)
	    {
	    case 0x00:
	      fputs(f00[(*code >> 3) & 7], out);
	      putc(' ', out);
	      code = mod_rm(out, code, seg, wreg, addr32);
	      break;
	    case 0x01:
	      fputs(f01[(*code >> 3) & 7], out);
	      putc(' ', out);
	      code = mod_rm(out, code, seg, wreg, addr32);
	      break;
	    case 0x02:
	      fputs("lar ", out);
	      code = mod_rm(out, code, seg, wreg, addr32);
	      break;
	    case 0x03:
	      fputs("lsl ", out);
	      code = mod_rm(out, code, seg, wreg, addr32);
	      break;
	    case 0x06:
	      fputs("clts", out);
	      break;
	    case 0x08:
	      fputs("invd", out);
	      break;
	    case 0x09:
	      fputs("wbinvd", out);
	      break;
	    case 0x10:
	      fputs("invlpg ", out);
	      code = mod_rm(out, code, seg, wreg, addr32);
	      break;
	    case 0x20:
	      fputs("mov ", out);
	      code = mod_reg_rm_r(out, code, seg, cr, reg32, addr32);
	      break;
	    case 0x21:
	      fputs("mov ", out);
	      code = mod_reg_rm_r(out, code, seg, dr, reg32, addr32);
	      break;
	    case 0x22:
	      fputs("mov ", out);
	      code = mod_reg_rm(out, code, seg, cr, reg32, addr32);
	      break;
	    case 0x23:
	      fputs("mov ", out);
	      code = mod_reg_rm(out, code, seg, dr, reg32, addr32);
	      break;
	    case 0x24:
	      fputs("mov ", out);
	      code = mod_reg_rm_r(out, code, seg, tr, reg32, addr32);
	      break;
	    case 0x26:
	      fputs("mov ", out);
	      code = mod_reg_rm(out, code, seg, tr, reg32, addr32);
	      break;
	    case 0x80:
	    case 0x81:
	    case 0x82:
	    case 0x83:
	    case 0x84:
	    case 0x85:
	    case 0x86:
	    case 0x87:
	    case 0x88:
	    case 0x89:
	    case 0x8a:
	    case 0x8b:
	    case 0x8c:
	    case 0x8d:
	    case 0x8e:
	    case 0x8f:
	      fprintf(out, "j%s %04hx", conditions[opcode & 15],
		      code - code0 + 2 + (short)(code[0] + (code[1] << 8)));
	      code += 2;
	      break;
	    case 0x90:
	    case 0x91:
	    case 0x92:
	    case 0x93:
	    case 0x94:
	    case 0x95:
	    case 0x96:
	    case 0x97:
	    case 0x98:
	    case 0x99:
	    case 0x9a:
	    case 0x9b:
	    case 0x9c:
	    case 0x9d:
	    case 0x9e:
	    case 0x9f:
	      fprintf(out, "set%s ", conditions[opcode & 15]);
	      code = mod_rm(out, code, seg, reg8, addr32);
	      break;

	    case 0xa0:
	      fputs("push fs", out);
	      break;

	    case 0xa1:
	      fputs("pop fs", out);
	      break;

	    case 0xa3:
	      fputs("bt ", out);
	      code = mod_reg_rm_r(out, code, seg, wreg, wreg, addr32);
	      break;

	    case 0xa4:
	      fputs("shld ", out);
	      mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	      fprintf(out, ",%d", *code++);
	      break;

	    case 0xa5:
	      fputs("shrd ", out);
	      mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	      fprintf(out, ",%d", *code++);
	      break;

	    case 0xa8:
	      fputs("push gs", out);
	      break;

	    case 0xa9:
	      fputs("pop gs", out);
	      break;

	    case 0xab:
	      fputs("bts ", out);
	      code = mod_reg_rm_r(out, code, seg, wreg, wreg, addr32);
	      break;

	    case 0xac:
	      fputs("shld ", out);
	      mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	      fputs(",cl", out);
	      break;

	    case 0xad:
	      fputs("shrd ", out);
	      mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	      fputs(",cl", out);
	      break;

	    case 0xb2:
	      fputs("lss ", out);
	      code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	      break;

	    case 0xb3:
	      fputs("btr ", out);
	      code = mod_reg_rm_r(out, code, seg, wreg, wreg, addr32);
	      break;

	    case 0xb4:
	      fputs("lfs ", out);
	      code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	      break;

	    case 0xb5:
	      fputs("lgs ", out);
	      code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	      break;

	    case 0xba:
	      fputs("bt? ", out);
	      code = mod_rm(out, code, seg, wreg, addr32);
	      fprintf(out, ",%d", *code++);
	      break;

	    case 0xbb:
	      fputs("btc ", out);
	      code = mod_reg_rm_r(out, code, seg, wreg, wreg, addr32);
	      break;

	    case 0xbc:
	      fputs("bsf ", out);
	      code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	      break;

	    case 0xbd:
	      fputs("bsr ", out);
	      code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	      break;

	    case 0xc8:
	    case 0xc9:
	    case 0xca:
	    case 0xcb:
	    case 0xcc:
	    case 0xcd:
	    case 0xce:
	    case 0xcf:
	      fprintf(out, "bswap %s", reg32[opcode & 7]);
	      break;

	    default:
	      fprintf(out, "ESC 0f %02x", opcode);
	      break;
	    }
	  break;

	case 0x16:
	  fputs("push ss", out);
	  break;

	case 0x17:
	  fputs("pop ss", out);
	  break;

	case 0x1e:
	  fputs("push ds", out);
	  break;

	case 0x1f:
	  fputs("pop ds", out);
	  break;

	case 0x26:
	  seg = "es:";
	  prefix = 1;
	  continue;

	case 0x27:
	  fputs("daa", out);
	  break;

	case 0x2e:
	  seg = "cs:";
	  prefix = 1;
	  continue;

	case 0x2f:
	  fputs("das", out);
	  break;

	case 0x36:
	  seg = "ss:";
	  prefix = 1;
	  continue;

	case 0x37:
	  fputs("aaa", out);
	  break;

	case 0x3e:
	  seg = "ds:";
	  prefix = 1;
	  continue;

	case 0x3f:
	  fputs("aas", out);
	  break;

	case 0x40:
	case 0x41:
	case 0x42:
	case 0x43:
	case 0x44:
	case 0x45:
	case 0x46:
	case 0x47:
	  fprintf(out, "inc %s", wreg[opcode & 7]);
	  break;

	case 0x48:
	case 0x49:
	case 0x4a:
	case 0x4b:
	case 0x4c:
	case 0x4d:
	case 0x4e:
	case 0x4f:
	  fprintf(out, "dec %s", wreg[opcode & 7]);
	  break;

	case 0x50:
	case 0x51:
	case 0x52:
	case 0x53:
	case 0x54:
	case 0x55:
	case 0x56:
	case 0x57:
	  fprintf(out, "push %s", wreg[opcode & 7]);
	  break;

	case 0x58:
	case 0x59:
	case 0x5a:
	case 0x5b:
	case 0x5c:
	case 0x5d:
	case 0x5e:
	case 0x5f:
	  fprintf(out, "pop %s", wreg[opcode & 7]);
	  break;

	case 0x60:
	  fputs("pusha", out);
	  break;

	case 0x61:
	  fputs("popa", out);
	  break;

	case 0x62:
	  fputs("bound ", out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0x63:
	  fputs("arpl ", out);
	  code = mod_reg_rm_r(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0x64:
	  seg = "fs:";
	  prefix = 1;
	  continue;

	case 0x65:
	  seg = "gs:";
	  prefix = 1;
	  continue;

	case 0x66:
	  data32 = !data32;
	  prefix = 1;
	  continue;

	case 0x67:
	  addr32 = !addr32;
	  prefix = 1;
	  continue;

	case 0x68:
	  if (data32)
	    {
	      fprintf(out, "push %d", immed32(code));
	      code += 4;
	    }
	  else
	    {
	      fprintf(out, "push %d", immed16(code));
	      code += 2;
	    }
	  break;

	case 0x69:
	  fputs("imul ", out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  if (data32)
	    {
	      fprintf(out, ",%d", immed32(code));
	      code += 4;
	    }
	  else
	    {
	      fprintf(out, ",%hd", immed16(code));
	      code += 2;
	    }
	  break;

	case 0x6a:
	  fprintf(out, "push %d", (signed char)*code++);
	  break;

	case 0x6b:
	  fputs("imul ", out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  fprintf(out, ",%d", (signed char)*code++);
	  break;

	case 0x6c:
	  fprintf(out, "insb es:[%s]", addr32 ? "edi" : "di");
	  break;

	case 0x6d:
	  fprintf(out, "ins%c es:[%s]", data32 ? 'd' : 'w',
		  addr32 ? "edi" : "di");
	  break;

	case 0x6e:
	  fprintf(out, "outsb dx,%s[%s]", seg, addr32 ? "esi" : "si");
	  break;

	case 0x6f:
	  fprintf(out, "outs%c dx,%s[%s]", data32 ? 'd' : 'w',
		  seg, addr32 ? "esi" : "si");
	  break;

	case 0x70:
	case 0x71:
	case 0x72:
	case 0x73:
	case 0x74:
	case 0x75:
	case 0x76:
	case 0x77:
	case 0x78:
	case 0x79:
	case 0x7a:
	case 0x7b:
	case 0x7c:
	case 0x7d:
	case 0x7e:
	case 0x7f:
	  fprintf(out, "j%s %04hx", conditions[opcode & 15],
		  code - code0 + (signed char)*code + 1);
	  code++;
	  break;

	case 0x80:
	case 0x82:
	  fputs(arith[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, reg8, addr32);
	  fprintf(out, ",%d", (signed char)*code++);
	  break;

	case 0x81:
	  fputs(arith[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, wreg, addr32);
	  if (data32)
	    {
	      fprintf(out, ",0x%x", immed32(code));
	      code += 4;
	    }
	  else
	    {
	      fprintf(out, ",0x%hx", immed16(code));
	      code += 2;
	    }
	  break;

	case 0x83:
	  fputs(arith[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, wreg, addr32);
	  fprintf(out, ",%d", (signed char)*code++);
	  break;

	case 0x84:
	  fputs("test ", out);
	  code = mod_reg_rm(out, code, seg, reg8, reg8, addr32);
	  break;

	case 0x85:
	  fputs("test ", out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0x86:
	  fputs("xchg ", out);
	  code = mod_reg_rm(out, code, seg, reg8, reg8, addr32);
	  break;

	case 0x87:
	  fputs("xchg ", out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0x88:
	  fputs("mov ", out);
	  code = mod_reg_rm_r(out, code, seg, reg8, reg8, addr32);
	  break;

	case 0x89:
	  fputs("mov ", out);
	  code = mod_reg_rm_r(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0x8a:
	  fputs("mov ", out);
	  code = mod_reg_rm(out, code, seg, reg8, reg8, addr32);
	  break;

	case 0x8b:
	  fputs("mov ", out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0x8c:
	  fputs("mov ", out);
	  code = mod_reg_rm_r(out, code, seg, sreg, reg16, addr32);
	  break;

	case 0x8d:
	  fputs("lea ", out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0x8e:
	  fputs("mov ", out);
	  code = mod_reg_rm(out, code, seg, sreg, reg16, addr32);
	  break;

	case 0x8f:
	  fputs("pop ", out);
	  code = mod_rm(out, code, seg, wreg, addr32);
	  break;

	case 0x90:
	  fputs("nop", out);
	  break;

	case 0x91:
	case 0x92:
	case 0x93:
	case 0x94:
	case 0x95:
	case 0x96:
	case 0x97:
	  fprintf(out, "xchg ax,%s", wreg[opcode & 7]);
	  break;

	case 0x98:
	  fputs("cbw", out);
	  break;

	case 0x99:
	  fputs("cwd", out);
	  break;

	case 0x9a:
	  fprintf(out, "call %04hx:%04hx", code[2] + (code[3] << 8),
		  code[0] + (code[1] << 8));
	  code += 4;
	  break;

	case 0x9b:
	  fputs("wait", out);
	  break;

	case 0x9c:
	  fputs("pushf", out);
	  break;

	case 0x9d:
	  fputs("popf", out);
	  break;

	case 0x9e:
	  fputs("sahf", out);
	  break;

	case 0x9f:
	  fputs("lahf", out);
	  break;

	case 0xa0:
	  fprintf(out, "mov al,%s[%04hx]", seg, immed16(code));
	  code += 2;
	  break;

	case 0xa1:
	  fprintf(out, "mov ax,%s[%04hx]", seg, immed16(code));
	  code += 2;
	  break;

	case 0xa2:
	  fprintf(out, "mov %s[%04hx],al", seg, immed16(code));
	  code += 2;
	  break;

	case 0xa3:
	  fprintf(out, "mov %s[%04hx],ax", seg, immed16(code));
	  code += 2;
	  break;

	case 0xa4:
	  fprintf(out, "movsb es:[%s],%s[%s]", addr32 ? "edi" : "di",
		  seg, addr32 ? "esi" : "si");
	  break;

	case 0xa5:
	  fprintf(out, "movs%c es:[%s],%s[%s]", data32 ? 'd' : 'w',
		  addr32 ? "edi" : "di", seg, addr32 ? "esi" : "si");
	  break;

	case 0xa6:
	  fputs("cmpsb", out);
	  break;

	case 0xa7:
	  if (data32)
	    fputs("cmpsd", out);
	  else
	    fputs("cmpsw", out);
	  break;

	case 0xa8:
	  fprintf(out, "test al,%d", *code++);
	  break;

	case 0xa9:
	  if (data32)
	    {
	      fprintf(out, "test eax,0x%x", immed32(code));
	      code += 4;
	    }
	  else
	    {
	      fprintf(out, "test ax,0x%hx", immed16(code));
	      code += 2;
	    }
	  break;

	case 0xaa:
	  fprintf(out, "stosb es:[%s]", addr32 ? "edi" : "di");
	  break;

	case 0xab:
	  fprintf(out, "stos%c es:[%s]", data32 ? 'd' : 'w',
		  addr32 ? "edi" : "di");
	  break;

	case 0xac:
	  fprintf(out, "lodsb %s[si]", seg);
	  break;

	case 0xad:
	  fprintf(out, "lods%c %s[%s]", data32 ? 'd' : 'c', seg,
		  addr32 ? "esi" : "si");
	  break;

	case 0xae:
	  fputs("scasb", out);
	  break;

	case 0xaf:
	  if (data32)
	    fputs("scasd", out);
	  else
	    fputs("scasw", out);
	  break;

	case 0xb0:
	case 0xb1:
	case 0xb2:
	case 0xb3:
	case 0xb4:
	case 0xb5:
	case 0xb6:
	case 0xb7:
	  fprintf(out, "mov %s,0x%x", reg8[opcode & 7], *code++);
	  break;

	case 0xb8:
	case 0xb9:
	case 0xba:
	case 0xbb:
	case 0xbc:
	case 0xbd:
	case 0xbe:
	case 0xbf:
	  if (data32)
	    {
	      fprintf(out, "mov %s,0x%x", wreg[opcode & 7], immed32(code));
	      code += 4;
	    }
	  else
	    {
	      fprintf(out, "mov %s,0x%hx", wreg[opcode & 7], immed16(code));
	      code += 2;
	    }
	  break;

	case 0xc0:
	  fputs(shift[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, reg8, addr32);
	  fprintf(out, ",%d", 31 & *code++);
	  break;

	case 0xc1:
	  fputs(shift[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, wreg, addr32);
	  fprintf(out, ",%d", 31 & *code++);
	  break;

	case 0xc2:
	  fprintf(out, "ret %d", *code++);
	  break;

	case 0xc3:
	  fputs("ret", out);
	  break;

	case 0xc4:
	  fputs("les ", out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0xc5:
	  fputs("lds ", out);
	  code = mod_reg_rm(out, code, seg, wreg, wreg, addr32);
	  break;

	case 0xc6:
	  fputs("mov ", out);
	  code = mod_rm(out, code, seg, reg8, addr32);
	  fprintf(out, ",%d", (signed char)*code++);
	  break;

	case 0xc7:
	  fputs("mov ", out);
	  code = mod_rm(out, code, seg, wreg, addr32);
	  if (data32)
	    {
	      fprintf(out, ",0x%x", immed32(code));
	      code += 4;
	    }
	  else
	    {
	      fprintf(out, ",%d", immed16(code));
	      code += 4;
	    }
	  break;

	case 0xc8:
	  fputs("enter", out);
	  break;

	case 0xc9:
	  fputs("leave", out);
	  break;

	case 0xca:
	  fprintf(out, "retf %d", immed16(code));
	  code += 2;
	  break;

	case 0xcb:
	  fputs("retf", out);
	  break;

	case 0xcc:
	  fputs("int 3", out);
	  break;

	case 0xcd:
	  fprintf(out, "int %d", *code++);
	  break;

	case 0xce:
	  fputs("into", out);
	  break;

	case 0xcf:
	  fputs("iret", out);
	  break;

	case 0xd0:
	  fputs(shift[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, reg8, addr32);
	  fputs(",1", out);
	  break;

	case 0xd1:
	  fputs(shift[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, wreg, addr32);
	  fputs(",1", out);
	  break;

	case 0xd2:
	  fputs(shift[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, reg8, addr32);
	  fputs(",cl", out);
	  break;

	case 0xd3:
	  fputs(shift[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, wreg, addr32);
	  fputs(",cl", out);
	  break;

	case 0xd4:
	  fprintf(out, "aam %d", *code++);
	  break;

	case 0xd5:
	  fprintf(out, "aad %d", *code++);
	  break;

	case 0xd7:
	  fputs("xlatb", out);
	  break;

	case 0xd8:
	  fputs(farith[(*code >> 3) & 7], out);
	  fputs("s ", out);
	  code = mod_rm(out, code, seg, freg, addr32);
	  break;
	case 0xda:
	  fputs(farith[(*code >> 3) & 7], out);
	  fputs("w ", out);
	  code = mod_rm(out, code, seg, freg, addr32);
	  break;
	case 0xdc:
	  fputs(farith[(*code >> 3) & 7], out);
	  fputs("l ", out);
	  code = mod_rm(out, code, seg, freg, addr32);
	  break;
	case 0xde:
	  fputs(farith[(*code >> 3) & 7], out);
	  fputs("i ", out);
	  code = mod_rm(out, code, seg, freg, addr32);
	  break;
	case 0xd9:
	case 0xdb:
	case 0xdd:
	case 0xdf:
	  fprintf(out, "esc %d,%d ", opcode & 7, ((*code >> 3) & 7));
	  if (freg_op[opcode & 7] & 1 << ((*code >> 3) & 7))
	    code = mod_rm(out, code, seg, freg, addr32);
	  else
	    code = mod_rm(out, code, seg, fop, addr32);
	  break;

	case 0xe0:
	  fprintf(out, "loopne %04hx", code - code0 + (signed char)*code + 1);
	  code++;
	  break;

	case 0xe1:
	  fprintf(out, "loope %04hx", code - code0 + (signed char)*code + 1);
	  code++;
	  break;

	case 0xe2:
	  fprintf(out, "loop %04hx", code - code0 + (signed char)*code + 1);
	  code++;
	  break;

	case 0xe3:
	  fprintf(out, "jcxz %04hx", code - code0 + (signed char)*code + 1);
	  code++;
	  break;

	case 0xe4:
	  fprintf(out, "in al,0x%hx", *code++);
	  break;

	case 0xe5:
	  fprintf(out, "in ax,0x%hx", *code++);
	  break;

	case 0xe6:
	  fprintf(out, "out 0x%hx,al", *code++);
	  break;

	case 0xe7:
	  fprintf(out, "out 0x%hx,ax", *code++);
	  break;

	case 0xe8:
	  fprintf(out, "call %04hx",
		  (unsigned short)(code - code0 + 2 + code[0] + (code[1] << 8)));
	  code += 2;
	  break;

	case 0xe9:
	  fprintf(out, "jmp %04hx",
		  (unsigned short)(code - code0 + 2 + code[0] + (code[1] << 8)));
	  code += 2;
	  break;

	case 0xea:
	  fprintf(out, "jmp %04hx:%04hx", code[2] + (code[3] << 8),
		  code[0] + (code[1] << 8));
	  code += 4;
	  break;

	case 0xeb:
	  fprintf(out, "jmp %04hx", code - code0 + (signed char)*code + 1);
	  code++;
	  break;

	case 0xec:
	  fputs("in al,dx", out);
	  break;

	case 0xed:
	  fputs("in ax,dx", out);
	  break;

	case 0xee:
	  fputs("out dx,al", out);
	  break;

	case 0xef:
	  fputs("out dx,ax", out);
	  break;

	case 0xf0:
	  fputs("lock ", out);
	  prefix = 1;
	  break;

	case 0xf2:
	  fputs("rep ", out);
	  prefix = 1;
	  continue;

	case 0xf3:
	  fputs("repe ", out);
	  prefix = 1;
	  continue;

	case 0xf4:
	  fputs("hlt", out);
	  break;

	case 0xf5:
	  fputs("cmc", out);
	  break;

	case 0xf6:
	  opcode = *code;
	  fputs(f7_code[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, reg8, addr32);
	  if (((opcode >> 3) & 7) == 0)
	    fprintf(out, ",%d", *code++);
	  break;

	case 0xf7:
	  opcode = *code;
	  fputs(f7_code[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, wreg, addr32);
	  if (((opcode >> 3) & 7) == 0)
	    {
	      if (data32)
		{
		  fprintf(out, ",%d", immed32(code));
		  code += 4;
		}
	      else
		{
		  fprintf(out, ",%d", immed16(code));
		  code += 4;
		}
	    }
	  break;

	case 0xf8:
	  fputs("clc", out);
	  break;

	case 0xf9:
	  fputs("stc", out);
	  break;

	case 0xfa:
	  fputs("cli", out);
	  break;

	case 0xfb:
	  fputs("sti", out);
	  break;

	case 0xfc:
	  fputs("cld", out);
	  break;

	case 0xfd:
	  fputs("std", out);
	  break;

	case 0xfe:
	  fputs(ff_code[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, reg8, addr32);
	  break;

	case 0xff:
	  fputs(ff_code[(*code >> 3) & 7], out);
	  putc(' ', out);
	  code = mod_rm(out, code, seg, wreg, addr32);
	  break;

	default:
	  fprintf(out, "%x", opcode);
	  break;
	}
      seg = "";
      data32 = def_size;
      addr32 = def_size;
      prefix = 0;
      putc('\n', out);
    }
}
